# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, List, Optional, Union

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import allow_single_or_multiple_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class ModuloLayer(BaseLayer):
    """
    Performs the modulo(x, y) operation on a given input tensor.
    If divisor is not set, inputs are assumed to be a list of two tensors and the
    first tensor is modulo'd by the second.
    If divisor is set, inputs must be a tensor.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        input_dtype: Optional[str] = None,
        output_dtype: Optional[str] = None,
        divisor: Optional[float] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initializes the ModuloLayer layer

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param divisor: The divisor to modulo the input by, defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.divisor = divisor

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [
            tf.int8,
            tf.int16,
            tf.int32,
            tf.int64,
            tf.uint8,
            tf.uint16,
            tf.uint32,
            tf.uint64,
            tf.bfloat16,
            tf.float16,
            tf.float32,
            tf.float64,
        ]

    @allow_single_or_multiple_tensor_input
    def _call(self, inputs: Union[Tensor, Iterable[Tensor]], **kwargs: Any) -> Tensor:
        """
        Performs the modulo(x, y) operation on either an iterable of input tensors or
        a single input tensor and a constant.

        Decorated with `@allow_single_or_multiple_tensor_input` to ensure that the input
        is either a single tensor or an iterable of tensors. Returns this result as a
        list of tensors for easier use here.

        :param inputs: Single tensor or iterable of tensors to perform the
        modulo(x, y) operation on.
        :returns: The tensor resulting from the modulo(x, y) operation.
        """
        if self.divisor is not None:
            if len(inputs) > 1:
                raise ValueError("If divisor is set, cannot have multiple inputs")
            cast_input, cast_divisor = self._force_cast_to_compatible_numeric_type(
                inputs[0], self.divisor
            )
            return tf.math.floormod(
                cast_input,
                cast_divisor,
            )
        else:
            if len(inputs) != 2:
                raise ValueError("If divisor is not set, must have exactly 2 inputs")
            return tf.math.floormod(inputs[0], inputs[1])

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the Modulo layer.
        Used for saving and loading from a model.

        Specifically adds the `divisor` to the config dictionary.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"divisor": self.divisor})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Literally a copy of pyspark.ml.util DefaultParamsWritable/Readable/Writer/Reader.
This is needed because Databricks runtimes > 13.3LTS have slow metadata writes because
they have introduced this v4.0.0 change: https://github.com/apache/spark/pull/47453
They have moved to using atomic writes for metadata, however this is intolerably slow
for large pipelines. We temporarily fix this back to the open source code.
This may need revisiting in the future."""
import os
from typing import Any, Dict, Optional

from pyspark import SparkContext
from pyspark.ml.param import Params
from pyspark.ml.util import (
    DefaultParamsReadable,
    DefaultParamsReader,
    DefaultParamsWritable,
    DefaultParamsWriter,
    MLWriter,
)


class KamaeDefaultParamsWriter(DefaultParamsWriter):
    """
    DefaultParamsWriter with a workaround for slow metadata writes in Databricks.
    Replicates the functionality of DefaultParamsWriter in PySpark 3.5.0 since
    Databricks uses different functionality
    """

    @staticmethod
    def saveMetadata(
        instance: "Params",
        path: str,
        sc: SparkContext,
        extraMetadata: Optional[Dict[str, Any]] = None,
        paramMap: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Saves metadata + Params to: path + "/metadata"
        - class
        - timestamp
        - sparkVersion
        - uid
        - paramMap
        - defaultParamMap (since 2.4.0)
        - (optionally, extra metadata)
        Parameters
        ----------
        extraMetadata : dict, optional
            Extra metadata to be saved at same level as uid, paramMap, etc.
        paramMap : dict, optional
            If given, this is saved in the "paramMap" field.
        """
        metadataPath = os.path.join(path, "metadata")
        metadataJson = DefaultParamsWriter._get_metadata_to_save(
            instance, sc, extraMetadata, paramMap
        )
        # This is the line we need to maintain.
        sc.parallelize([metadataJson], 1).saveAsTextFile(metadataPath)


class KamaeDefaultParamsWritable(DefaultParamsWritable):
    """
    DefaultParamsWritable with a workaround for slow metadata writes in Databricks.
    Replicates the functionality of DefaultParamsWritable in PySpark 3.5.0 since
    Databricks uses different functionality
    """

    def write(self) -> MLWriter:
        """Returns a DefaultParamsWriter instance for this class."""
        from pyspark.ml.param import Params

        if isinstance(self, Params):
            return KamaeDefaultParamsWriter(self)
        else:
            raise TypeError(
                """Cannot use KamaeDefaultParamsWritable with type %s because it does
                not extend Params.""",
                type(self),
            )


class KamaeDefaultParamsReader(DefaultParamsReader):
    """
    DefaultParamsReadable with a workaround for slow metadata writes in Databricks.
    Replicates the functionality of DefaultParamsReadable in PySpark 3.5.0 since
    Databricks uses different functionality
    """

    @staticmethod
    def loadMetadata(
        path: str, sc: SparkContext, expectedClassName: str = ""
    ) -> Dict[str, Any]:
        """
        Load metadata saved using :py:meth:`DefaultParamsWriter.saveMetadata`
        Parameters
        ----------
        path : str
        sc : :py:class:`pyspark.SparkContext`
        expectedClassName : str, optional
            If non empty, this is checked against the loaded metadata.
        """
        metadataPath = os.path.join(path, "metadata")
        # This is the line we need to maintain.
        metadataStr = sc.textFile(metadataPath, 1).first()
        loadedVals = DefaultParamsReader._parseMetaData(metadataStr, expectedClassName)
        return loadedVals


class KamaeDefaultParamsReadable(DefaultParamsReadable):
    """
    DefaultParamsReadable with a workaround for slow metadata writes in Databricks.
    Replicates the functionality of DefaultParamsReadable in PySpark 3.5.0 since
    Databricks uses different functionality
    """

    @classmethod
    def read(cls) -> "KamaeDefaultParamsReader":
        """Returns a KamaeDefaultParamsReader instance for this class."""
        return KamaeDefaultParamsReader(cls)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(kamae.__name__)
class StringAffixLayer(BaseLayer):
    """
    Performs a prefixing and suffing on the input tensor.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        input_dtype: Optional[str] = None,
        output_dtype: Optional[str] = None,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialises the String Affix layer.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param prefix: The prefix to apply to tensor.
        :param suffix: The suffix to apply to tensor.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.prefix = prefix
        self.suffix = suffix
        self.validate_params()

    def validate_params(self) -> None:
        """
        Validates the parameters of the layer.
        :raises ValueError: If both prefix and suffix are not set.
        """
        if (self.prefix is None or self.prefix == "") and (
            self.suffix is None or self.suffix == ""
        ):
            raise ValueError(
                "Either prefix or suffix must be set. Otherwise nothing to affix."
            )

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs: Any) -> Tensor:
        """
        Prefixes and suffixes a given input tensor.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Input tensor to affix. Must be string tensors.
        :returns: A tensor with affixed values - same shape as input.
        """
        x = inputs
        if self.prefix:
            x = tf.strings.join([self.prefix, x])
        if self.suffix:
            x = tf.strings.join([x, self.suffix])
        return x

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the StringAffix layer.
        Used for saving and loading from a model.

        Specifically adds the `prefix` and `suffix` values to the configuration.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"prefix": self.prefix, "suffix": self.suffix})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType, DoubleType, FloatType, StringType

from kamae.spark.params import (
    ListwiseStatisticsParams,
    MultiInputSingleOutputParams,
    NanFillValueParams,
    SingleInputSingleOutputParams,
)
from kamae.spark.utils import check_and_apply_listwise_op
from kamae.tensorflow.layers import ListMinLayer

from .base import BaseTransformer


class ListMinTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    MultiInputSingleOutputParams,
    ListwiseStatisticsParams,
    NanFillValueParams,
):
    """
    Calculate the listwise minimum across the query id column.
    - If inputCol is set, the transformer calculates the minimum of the input column
    based on all the items with the same query id column value.
    - If inputCols is set, behaviour depends on the value of withSegment:
        - If withSegment = True: the transformer calculates the minimum of the first column
        with the same query id column value, segmented by values of the second column.

        Example: calculate the minimum price of hotels within star ratings, in the same query.

        - If withSegment = False: the transformer calculates the minimum of the first column
        with the same query id column value, based on second column's topN items.
        When using the second input as sorting column, topN must be provided.
        By using the topN items to calculate the statistics, we can better approximate
        the real statistics in production. A large enough topN should be used, to obtain a
        good approximation of the statistics, and an important feature to sort on, such as
        item's production.

        Example: calculate the minimum price in the same query, based on the top N
        items sorted by descending production.

    :param inputCol: Value column, on which to calculate the minimum.
    :param inputCols: Input column names.
    - The first is the value column, on which to calculate the minimum.
    - The second is the sort or segment column. The role of the second input is governed
    by the value of withSegment as described above.
    :param outputCol: Name of output col.
    :param inputDtype: Data Type of input.
    :param outputDtype: Data Type of output.
    :param layerName: The name of the transformer, which typically
    should be the name of the produced feature.
    :param queryIdCol: Name of column to aggregate upon. It is required.
    :param topN: Filter for limiting the items to calculate the statistics. Not used when withSegment = True.
    :param sortOrder: Option of 'asc' or 'desc' which defines order
    for listwise operation. Default is 'asc'. Not used when withSegment = True.
    :param withSegment: Whether to use the second input column to partition the statistic
    calculation. Defaults to False.
    :param minFilterValue: Minimum value to remove padded values
    defaults to >= 0.
    :nanFillValue: Value to fill NaNs results with. Defaults to 0.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        queryIdCol: Optional[str] = None,
        topN: Optional[int] = None,
        sortOrder: str = "asc",
        withSegment: bool = False,
        minFilterValue: Optional[float] = None,
        nanFillValue: float = 0.0,
    ) -> None:
        super().__init__()
        self._setDefault(
            topN=None,
            sortOrder="asc",
            minFilterValue=None,
            nanFillValue=0,
            withSegment=False,
        )
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [
            FloatType(),
            DoubleType(),
            StringType(),
        ]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Calculate the listwise minimum, optionally sorting and
        filtering based on the second input column.
        :param dataset: The dataframe with signals and features.
        :returns: The dataframe dataset with the new feature.
        """
        if not self.isDefined("queryIdCol"):
            raise ValueError("queryIdCol must be set on listwise transformers.")

        # Define the columns to use for the calculation
        if self.isDefined("inputCols"):
            with_segment = self.getWithSegment()
            if with_segment:
                val_col_name = self.getInputCols()[0]
                segment_col_name = self.getInputCols()[1]
                sort_col_name = None
            else:
                val_col_name = self.getInputCols()[0]
                sort_col_name = self.getInputCols()[1]
                segment_col_name = None
        else:
            val_col_name = self.getInputCol()
            sort_col_name = None
            segment_col_name = None

        dataset = dataset.withColumn(
            self.getOutputCol(),
            check_and_apply_listwise_op(
                dataset,
                F.min,
                self.getQueryIdCol(),
                val_col_name,
                sort_col_name,
                segment_col_name,
                self.getSortOrder(),
                self.getTopN(),
                self.getMinFilterValue(),
            ),
        )

        # Replace Nulls/Nans
        dataset = dataset.fillna({self.getOutputCol(): self.getNanFillValue()})

        return dataset

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the listwise-minimum transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs an averaging operation.
        """
        return ListMinLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            top_n=self.getTopN(),
            sort_order=self.getSortOrder(),
            with_segment=self.getWithSegment(),
            min_filter_value=self.getMinFilterValue(),
            nan_fill_value=self.getNanFillValue(),
            axis=1,
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(kamae.__name__)
class ArraySplitLayer(BaseLayer):
    """
    Performs a splitting of the input tensor into a list of tensors.
    Expands dimensions to ensure the output tensors are the same shape as the input.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        input_dtype: Optional[str] = None,
        output_dtype: Optional[str] = None,
        axis: int = -1,
        **kwargs: Any,
    ) -> None:
        """
        Initialises the ArraySplitLayer layer.

        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param axis: Axis to split on. Defaults to -1.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.axis = axis

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return None

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs: Any) -> List[Tensor]:
        """
        Splits the input tensor along the specified axis.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if an iterable of tensors is passed
        in.

        :param inputs: Tensor to split.
        :returns: List of split tensors.
        """
        return [
            tf.expand_dims(y, axis=self.axis)
            for y in tf.unstack(inputs, axis=self.axis)
        ]

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the VectorSplit layer.
        Used for saving and loading from a model.

        Specifically, adds the `axis` to the config.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"axis": self.axis})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class StringListToStringLayer(BaseLayer):
    """
    A layer that converts a list of strings to a single string along the specified
    axis.
    If `keepdims` is `True`, the shape is retained.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        input_dtype: Optional[str] = None,
        output_dtype: Optional[str] = None,
        axis: int = -1,
        separator: str = "",
        keepdims: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Initialises the StringListToStringLayer layer.

        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param axis: The axis along which to join the strings. Defaults to `-1`.
        :param separator: The separator to use when joining the strings.
        Defaults to `""`.
        :param keepdims: Whether to keep the shape of the input tensor. Defaults to
        `False`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.axis = axis
        self.separator = separator
        self.keepdims = keepdims

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs: Any) -> Tensor:
        """
        Joins the strings along the specified axis with the specified separator.
        If `keepdims` is `True`, the shape is retained. Otherwise the shape is
        reduced along the specified axis.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if an iterable of tensors is passed
        in.

        :param inputs: Input tensor.
        :returns: Tensor with strings joined along the specified axis.
        """
        return tf.strings.reduce_join(
            inputs, axis=self.axis, separator=self.separator, keepdims=self.keepdims
        )

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the StringListToString layer.
        Used for saving and loading from a model.

        Specifically adds the `axis`, `separator` and `keepdims` to the config
        dictionary.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "axis": self.axis,
                "separator": self.separator,
                "keepdims": self.keepdims,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input, listify_tensors

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class MinMaxScaleLayer(BaseLayer):
    """
    Performs a min-max scaling operation on the input tensor(s).
    This is used to standardize/transform the input tensor
    to the range [0, 1] using the minimum and maximum values.

    Formula: (x - min)/(max - min)
    """

    def __init__(
        self,
        min: Union[List[float], np.array],
        max: Union[List[float], np.array],
        name: Optional[str] = None,
        input_dtype: Optional[str] = None,
        output_dtype: Optional[str] = None,
        axis: int = -1,
        mask_value: Optional[float] = None,
        **kwargs: Any,
    ) -> None:
        """
        Intialise the MinMaxScaleLayer layer.
        :param min: The min value(s) to use during scaling.
        :param max: The max value(s) to use during scaling.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param axis: The axis that should have a separate min and max. For
        example, if shape is `(None, 5)` and `axis=1`, the layer will track 5
        separate min and max values for the last axis.
        :param mask_value: Value which should be ignored during scaling.
        """
        super().__init__(
            name=name,
            input_dtype=input_dtype,
            output_dtype=output_dtype,
            **kwargs,
        )  # Standardize `axis` to a tuple.
        if axis is None:
            axis = ()
        elif isinstance(axis, int):
            axis = (axis,)
        else:
            axis = tuple(axis)

        self.axis = axis
        self.input_min = min
        self.input_max = max
        self.mask_value = mask_value

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.bfloat16, tf.float16, tf.float32, tf.float64]

    def build(self, input_shape: Tuple[int]) -> None:
        """
        Builds shapes for the min and max tensors.

        Specifically, understands which axis to compute the scaling across
        and broadcasts the min and max tensors to match the input shape.

        :param input_shape: The shape of the input tensor.
        :returns: None - layer is built.
        """
        super().build(input_shape)

        if isinstance(input_shape, (list, tuple)) and all(
            isinstance(shape, (tf.TensorShape, list, tuple)) for shape in input_shape
        ):
            # This seems to be needed to handle sending in multiple inputs as a list.
            # Although this layer should only have one input, so this is a bit of a
            # hack. We catch this nicely in call method with a decorator. Maybe we
            # should do the same here?
            input_shape = input_shape[0]

        input_shape = tf.TensorShape(input_shape).as_list()
        ndim = len(input_shape)
        self._build_input_shape = input_shape

        if any(a < -ndim or a >= ndim for a in self.axis):
            raise ValueError(
                f"""All `axis` values must be in the range [-ndim, ndim). "
                Found ndim: `{ndim}`, axis: {self.axis}"""
            )

        # Axes to be kept, replacing negative values with positive equivalents.
        # Sorted to avoid transposing axes.
        keep_axis = sorted([d if d >= 0 else d + ndim for d in self.axis])
        # All axes to be kept should have known shape.
        for d in keep_axis:
            if input_shape[d] is None:
                raise ValueError(
                    f"""All `axis` values to be kept must have known shape. "
                    Got axis: {self.axis},
                    input shape: {input_shape}, with unknown axis at index: {d}"""
                )
        # Broadcast any reduced axes.
        broadcast_shape = [input_shape[d] if d in keep_axis else 1 for d in range(ndim)]
        min_and_max_shape = tuple(input_shape[d] for d in keep_axis)
        min_tensor = self.input_min * np.ones(min_and_max_shape)
        max_tensor = self.input_max * np.ones(min_and_max_shape)
        self.min = tf.reshape(min_tensor, broadcast_shape)
        self.max = tf.reshape(max_tensor, broadcast_shape)

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the MinMaxScaleLayer layer.
        Used for saving and loading from a model.
        Specifically adds additional parameters to the base configuration.
        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        # Ensure mean and variance are lists for serialization.
        config.update(
            {
                "min": listify_tensors(self.input_min),
                "max": listify_tensors(self.input_max),
                "axis": self.axis,
            }
        )
        return config

    def get_build_config(self) -> Optional[Dict[str, Any]]:
        """
        Gets the build configuration of the MinMaxScaleLayer layer.

        Used for saving and loading from a model.

        :returns: Dictionary of the build configuration of the layer.
        """
        if self._build_input_shape:
            return {"input_shape": self._build_input_shape}

    def build_from_config(self, config: Dict[str, Any]) -> None:
        """
        Builds the min/max tensor shapes from the provided configuration.

        Specifically it calls the `build` method with the input shape in order to
        construct the min and max tensors with the correct shape.

        :param config: Configuration dictionary containing the input shape.
        :returns: None - layer is built.
        """
        if config:
            self.build(config["input_shape"])

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs: Any) -> Tensor:
        """
        Performs normalization on the input tensor(s) to scale it to the range [0, 1]
        Decorated with `@enforce_single_tensor_input` to ensure that
        the input is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.
        :param inputs: Input tensor to perform the normalization on.
        :returns: The input tensor with the normalization applied.
        """
        # Ensure min and max match input dtype.
        min_tensor = self._cast(self.min, inputs.dtype.name)
        max_tensor = self._cast(self.max, inputs.dtype.name)
        normalized_outputs = tf.math.divide_no_nan(
            tf.math.subtract(inputs, min_tensor),
            tf.math.subtract(max_tensor, min_tensor),
        )
        if self.mask_value is not None:
            mask = tf.equal(inputs, self.mask_value)
            normalized_outputs = tf.where(
                mask, inputs, self._cast(normalized_outputs, inputs.dtype.name)
            )
        return normalized_outputs



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class BucketizeLayer(BaseLayer):
    """
    Performs a bucketing operation on the input tensor.
    Given a list of splits, the input tensor is bucketed into
    the corresponding bucket. For example, if the splits are
    [0, 1, 2, 3], then the input tensor is bucketed into 4 buckets:
    (-inf, 0), [0, 1), [1, 2), [2, 3), [3, inf).
    These buckets are int64 values, starting from 1. The 0 index
    is reserved for padding values.
    """

    def __init__(
        self,
        splits: List[float],
        name: Optional[str] = None,
        input_dtype: Optional[str] = None,
        output_dtype: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialises the BucketizeLayer layer.

        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param splits: The splits to use for bucketing.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        if splits != sorted(splits):
            raise ValueError("`splits` argument must be a sorted list!")
        self.splits = splits

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.int32, tf.int64, tf.float32, tf.float64]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs: Any) -> Tensor:
        """
        Performs the bucketing operation on the input tensor.

        Decorated with `@enforce_single_tensor_input` to ensure that
        the input is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Input tensor to bucket.
        :returns: Bucketed tensor.
        """
        # We add 1 to the output of the bucket layer so that we can use
        # 0 index as a padding value.
        bucketed_outputs = tf.raw_ops.Bucketize(input=inputs, boundaries=self.splits)
        return self._cast(tf.math.add(bucketed_outputs, 1), "int64")

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the Bucketizer layer.
        Used for saving and loading from a model.

        Specifically adds the `splits` argument to the base config.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"splits": self.splits})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import allow_single_or_multiple_tensor_input, get_top_n

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class ListMedianLayer(BaseLayer):
    """
    Calculate the median across the axis dimension.
    - If one tensor is passed, the transformer calculates the median of the tensor
    based on all the items in the given axis dimension.
    - If inputCols is set, the transformer calculates the median of the first tensor
    based on second tensor's topN items in the same given axis dimension.

    By using the topN items to calculate the statistics, we can better approximate
    the real statistics in production. It is suggested to use a large enough topN to
    get a good approximation of the statistics, and an important feature to sort on,
    such as item's past production.

    Example: calculate the median price in the same query, based only on the top N
    items sorted by descending production.

    WARNING: ListMedianLayer requires at least rank 3 tensor input.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        input_dtype: Optional[str] = None,
        output_dtype: Optional[str] = None,
        top_n: Optional[int] = None,
        sort_order: str = "asc",
        min_filter_value: Optional[float] = None,
        nan_fill_value: float = 0.0,
        axis: int = 1,
        **kwargs: Any,
    ) -> None:
        """
        Initializes the Listwise Median layer.

        WARNING: The code is fully tested for axis=1 only. Further testing is needed.

        WARNING: The code can be affected by the value of the padding items. Always
        make sure to filter out the padding items value with min_filter_value.

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param top_n: The number of top items to consider when calculating the median.
        :param sort_order: The order to sort the second tensor by. Defaults to `asc`.
        :param min_filter_value: The minimum filter value to ignore values during
        calculation. Defaults to None (no filter).
        :param nan_fill_value: The value to fill NaNs results with. Defaults to 0.
        :param axis: The axis to calculate the statistics across. Defaults to 1.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.top_n = top_n
        self.sort_order = sort_order
        self.min_filter_value = min_filter_value
        self.nan_fill_value = nan_fill_value
        self.axis = axis

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [
            tf.bfloat16,
            tf.float16,
            tf.float32,
            tf.float64,
        ]

    def sort_with_nans_last(self, tensor: Tensor) -> Tensor:
        """
        Sorts a tensor while placing NaN values at the end along the specified axis.

        :param tensor: The input tensor.
        :param axis: The axis along which to sort.
        :returns: The sorted tensor with NaN values placed at the end.
        """
        # Replace NaNs with a very large value to move them to the end
        masked_tensor = tf.where(tf.math.is_nan(tensor), tensor.dtype.max, tensor)

        # Sort the tensor along the specified axis
        sorted_masked_tensor = tf.sort(masked_tensor, axis=self.axis)

        # Replace the very large values back with NaN after sorting
        sorted_masked_tensor = tf.where(
            tf.equal(sorted_masked_tensor, tensor.dtype.max),
            tf.constant(float("nan"), dtype=tensor.dtype),
            sorted_masked_tensor,
        )

        return sorted_masked_tensor

    @allow_single_or_multiple_tensor_input
    def _call(self, inputs: Iterable[Tensor], **kwargs: Any) -> Tensor:
        """
        Calculate the listwise median, optionally sorting and
        filtering based on the second input tensor.

        :param inputs: The iterable tensor for the feature.
        :returns: The new tensor result column.
        """
        val_tensor = inputs[0]
        output_shape = tf.shape(val_tensor)

        with_sort = True if len(inputs) == 2 else False
        sort_tensor = inputs[1] if with_sort else None

        if with_sort and self.top_n is None:
            raise ValueError("topN must be specified when using a sort column.")

        if with_sort:
            # Get the values corresponding to the top N item in the sort tensor
            filtered_tensor = get_top_n(
                val_tensor=val_tensor,
                axis=self.axis,
                sort_tensor=sort_tensor,
                sort_order=self.sort_order,
                top_n=self.top_n,
            )
        else:
            filtered_tensor = val_tensor

        # Assign nan to elements less than or equal to the threshold
        if self.min_filter_value is not None:
            filtered_tensor = tf.where(
                filtered_tensor >= self.min_filter_value,
                filtered_tensor,
                tf.constant(float("nan"), dtype=val_tensor.dtype),
            )
        else:
            filtered_tensor = filtered_tensor

        # Get the number of non-nan values
        num_valid_values = tf.reduce_sum(
            tf.cast(tf.math.is_finite(filtered_tensor), tf.int32), axis=self.axis
        )

        # Sort the values along the list dimension
        sorted_filtered_tensor = self.sort_with_nans_last(filtered_tensor)

        # Calculate the indices of the median values
        lower_index = (num_valid_values - 1) // 2
        upper_index = tf.minimum(lower_index + 1, num_valid_values - 1)

        # Gather the median values for each feature
        batch_size = tf.shape(filtered_tensor)[0]
        batch_indices = tf.range(batch_size)[:, tf.newaxis, tf.newaxis]
        lower_indices = tf.concat([batch_indices, lower_index[:, tf.newaxis]], axis=-1)
        lower_medians = tf.gather_nd(sorted_filtered_tensor, lower_indices)
        upper_indices = tf.concat([batch_indices, upper_index[:, tf.newaxis]], axis=-1)
        upper_medians = tf.gather_nd(sorted_filtered_tensor, upper_indices)

        # Calculate the average of lower and upper medians for even cases
        listwise_median = tf.where(
            tf.math.mod(num_valid_values[:, tf.newaxis], 2) == 0,
            (lower_medians + upper_medians) / 2.0,
            lower_medians,
        )

        # Fill nan
        is_integer = listwise_median.dtype.is_integer
        nan_val = int(self.nan_fill_value) if is_integer else self.nan_fill_value
        listwise_median = tf.where(
            tf.math.is_nan(listwise_median),
            tf.constant(nan_val, dtype=listwise_median.dtype),
            listwise_median,
        )

        # Broadcast the stat to each item in the list
        # WARNING: If filter creates empty items list, the result will be NaN
        listwise_median = tf.broadcast_to(listwise_median, output_shape)

        return listwise_median

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the layer.
        Used for saving and loading from a model.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "top_n": self.top_n,
                "sort_order": self.sort_order,
                "min_filter_value": self.min_filter_value,
                "nan_fill_value": self.nan_fill_value,
                "axis": self.axis,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import DataFrame
from pyspark.sql.types import ArrayType, DataType, IntegerType, StringType

from kamae.spark.params import MaskStringValueParams, SingleInputSingleOutputParams
from kamae.spark.utils import (
    min_hash_udf,
    single_input_single_output_array_udf_transform,
)
from kamae.tensorflow.layers import MinHashIndexLayer

from .base import BaseTransformer


class MinHashIndexParams(MaskStringValueParams):
    """
    Mixin class containing bin parameter needed for the MinHashIndexTransformer.
    """

    numPermutations = Param(
        Params._dummy(),
        "numPermutations",
        """Number of permutations to perform the min hashing.
        Will return an array with length equal to this.""",
        typeConverter=TypeConverters.toInt,
    )

    def setNumPermutations(self, value: int) -> "MinHashIndexParams":
        """
        Sets the numPermutations parameter.

        :param value: Integer value for the number of bins to use for hash indexing.
        :returns: Instance of class mixed in.
        """
        if value <= 0:
            raise ValueError("Number of permutations must be greater than 0.")
        return self._set(numPermutations=value)

    def getNumPermutations(self) -> int:
        """
        Gets the numPermutations parameter.

        :returns: Integer value for the number of bins to use for hash indexing.
        """
        return self.getOrDefault(self.numPermutations)


class MinHashIndexTransformer(
    BaseTransformer,
    MinHashIndexParams,
    SingleInputSingleOutputParams,
):
    """
    MinHash indexer Spark Transformer for use in Spark pipelines.
    This transformer hashes the input string set using the MinHash algorithm:
    https://en.wikipedia.org/wiki/MinHash

    MinHash approximates the Jaccard similarity between sets by hashing the elements of
    the sets and returning a fixed-length signature. This length is determined by the
    numPermutations parameter, which defaults to 128. The output is an array of integer
    bits.

    Setting the maskValue parameter allows you to ignore a specific value in the
    input column when computing the min hash. This is useful if you have padded arrays
    as then a padded array with the same unique elements as another non-padded array
    will be considered equal.

    NOTE: If your data contains null characters:
    https://en.wikipedia.org/wiki/Null_character
    This transformer could fail since the hashing algorithm used cannot accept null
    characters. If you have null characters in your data, you should remove them.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        numPermutations: int = 128,
        maskValue: Optional[str] = None,
    ) -> None:
        """
        Instantiates a MinHashIndexTransformer transformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param numPermutations: Number of permutations of your output min hash.
        Defaults to 128. This is the length of the output array.
        :param maskValue: Mask value to use when indexing the input column.
        If set, the mask value will be ignored when computing the min hash.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(numPermutations=128, maskValue=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column named outputCol with the
        min hash indexed input column.

        :param dataset: Pyspark DataFrame to transform.
        :returns: Transformed pyspark dataFrame.
        """
        if not self.isDefined("numPermutations"):
            raise ValueError("numPermutations parameter must be set.")
        num_permutations = self.getNumPermutations()
        mask_value = self.getMaskValue()

        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        if not isinstance(input_datatype, ArrayType):
            raise ValueError(
                f"""Input column {self.getInputCol()} must be of type ArrayType,
                but got {input_datatype}."""
            )
        output_col = single_input_single_output_array_udf_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: min_hash_udf(
                labels=x, num_permutations=num_permutations, mask_value=mask_value
            ),
            udf_return_element_datatype=IntegerType(),
        )
        return dataset.withColumn(
            self.getOutputCol(),
            output_col,
        )

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer that performs the min hash indexing.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
        that performs the hash indexing operation.
        """
        return MinHashIndexLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            num_permutations=self.getNumPermutations(),
            mask_value=self.getMaskValue(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import DataFrame, Window
from pyspark.sql.types import (
    ByteType,
    DataType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
)

from kamae.spark.params import ListwiseParams, SingleInputSingleOutputParams
from kamae.spark.utils import check_listwise_columns
from kamae.tensorflow.layers import ListRankLayer

from .base import BaseTransformer


class ListRankTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    ListwiseParams,
):
    """
    Calculate the listwise rank across the query id column.

    Example: calculate the rank of items within a query, given the score.

    :param inputCol: Value column, on which to calculate the rank.
    :param outputCol: Name of output col.
    :param inputDtype: Data Type of input.
    :param outputDtype: Data Type of output.
    :param layerName: The name of the transformer, which typically
    should be the name of the produced feature.
    :param queryIdCol: Name of column to aggregate upon. It is required.
    :param sortOrder: Option of 'asc' or 'desc' which defines order
    for listwise operation. Default is 'desc'.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        queryIdCol: Optional[str] = None,
        sortOrder: str = "desc",
    ) -> None:
        super().__init__()
        self._setDefault(sortOrder="desc")
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [
            FloatType(),
            DoubleType(),
            ByteType(),
            ShortType(),
            IntegerType(),
            LongType(),
        ]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Calculate the rank.
        :param dataset: The dataframe with signals and features.
        :returns: The dataframe dataset with the new feature.
        """
        if not self.isDefined("queryIdCol"):
            raise ValueError("queryIdCol must be set on listwise transformers.")

        # Get params
        input_col = self.getInputCol()
        query_id_col = self.getQueryIdCol()
        output_col = self.getOutputCol()
        sort_order = self.getSortOrder()

        # Validate listwise cols
        check_listwise_columns(
            dataset=dataset,
            query_col_name=query_id_col,
            value_col_name=input_col,
            sort_col_name=None,
        )

        # Set sort order
        if sort_order == "asc":
            sort_col = F.col(input_col).asc()
        elif sort_order == "desc":
            sort_col = F.col(input_col).desc()
        else:
            raise ValueError(f"Invalid sortOrder: {sort_order}")
        # Define window spec
        window_spec = Window.partitionBy(query_id_col).orderBy(sort_col)

        # Calculate the rank
        dataset = dataset.withColumn(output_col, F.row_number().over(window_spec))

        return dataset

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the listwise-rank transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs a ranking operation.
        """
        return ListRankLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            sort_order=self.getSortOrder(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf
from tensorflow.keras.layers import Hashing

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class MinHashIndexLayer(BaseLayer):
    """
    Performs min hashing of the input tensor as described here:
    https://en.wikipedia.org/wiki/MinHash

    MinHash approximates the Jaccard similarity between sets by hashing the elements of
    the sets and returning a fixed-length signature. This length is determined by the
    num_permutations parameter, which defaults to 128. The output is an array of integer
    bits.

    Setting the mask_value parameter allows you to ignore a specific value in the
    input column when computing the min hash. This is useful if you have padded arrays
    as then a padded array with the same unique elements as another non-padded array
    will be considered equal.

    The minimum is computed across the last dimension of the input tensor.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        input_dtype: Optional[str] = None,
        output_dtype: Optional[str] = None,
        num_permutations: int = 128,
        mask_value: Optional[str] = None,
        axis: int = -1,
        **kwargs: Any,
    ) -> None:
        """
        Initialises the MinHashIndexLayer layer.

        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param num_permutations: Number of permutations to use for the min hashing.
            Defaults to 128.
        :param mask_value: A value that represents masked inputs, which are ignored when
        computing the min hash. Defaults to None, meaning no mask term will be added.
        :param axis: The axis along which to compute the min hash.
        Defaults to -1 (last axis).
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.num_permutations = num_permutations
        self.axis = axis
        self.mask_value = mask_value
        self.hash_fn = Hashing(
            # Set the number of bins to the maximum integer value. We just want to hash
            # the input without binning it, so we use the maximum integer value.
            num_bins=tf.int32.max
        )

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs: Any) -> Tensor:
        """
        Performs the min hash indexing on the input tensor.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Input tensor to be encoded.
        :returns: Encoded tensor.
        """
        min_hash_signature = []
        for i in range(self.num_permutations):
            # Salt the input
            salted_inputs = tf.strings.join(
                [inputs, tf.zeros_like(inputs)], separator=str(i)
            )
            # Hash the salted inputs.
            if self.mask_value is not None:
                hashed_inputs = tf.where(
                    tf.equal(salted_inputs, f"{self.mask_value}{i}"),
                    # Use the maximum integer value for masked inputs, therefore it is
                    # never selected as the minimum.
                    tf.ones_like(salted_inputs, dtype=tf.int64) * tf.int32.max,
                    self.hash_fn(salted_inputs),
                )
            else:
                hashed_inputs = self.hash_fn(salted_inputs)
            min_hash_value = tf.reduce_min(hashed_inputs, axis=self.axis, keepdims=True)
            min_hash_bit = min_hash_value & 1
            min_hash_signature.append(min_hash_bit)

        # Concatenate the min hash values to form the final signature.
        return tf.concat(min_hash_signature, axis=self.axis)

    def get_config(self) -> Dict[str, Any]:
        """
        Returns the configuration of the MinHashIndex layer.

        :returns: Configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "num_permutations": self.num_permutations,
                "mask_value": self.mask_value,
                "axis": self.axis,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Optional, Union

import tensorflow as tf

from kamae.tensorflow.typing import Tensor


def map_fn_w_axis(
    elems: Union[Tensor, List[Tensor]],
    fn: Callable[[Tensor], Tensor],
    fn_output_signature: tf.dtypes.DType,
    axis: int = -1,
    parallel_iterations: Optional[int] = None,
    swap_memory: bool = False,
    infer_shape: bool = True,
    name: Optional[str] = None,
) -> Tensor:
    """
    Applies a function to a specific axis of a tensor using map_fn.
    Specifically uses `tf.transpose` and `tf.reshape` to rearrange the tensor so that
    the specified axis is preserved, the tensor is 2D and thus can be used with map_fn.

    After applying map_fn, the tensor is reshaped and transposed back to the original
    shape.

    :param elems: The input tensor or list of tensors.
    :param fn: The function to apply to the tensor. Must take a single tensor as input
    and return a tensor.
    :param fn_output_signature: The output signature of the function.
    :param axis: The axis to apply the function to. Defaults to -1.
    :param parallel_iterations: The number of iterations to run in parallel. Defaults to
    None.
    :param swap_memory: Whether to use memory swapping. Defaults to False.
    :param infer_shape: Whether to infer the shape of the output. Defaults to True.
    :param name: The name of the operation. Defaults to None.
    """

    def apply_transpose_and_reshape(tensor: Tensor) -> Tensor:
        transposed = tf.transpose(tensor, perm=transpose_perm)
        reshaped = tf.reshape(transposed, tf.stack([-1, tf.shape(tensor)[axis]]))
        return reshaped

    def apply_undo_transpose_and_reshape(
        output: Tensor, transposed_shape: Tensor, identity_perm: Tensor, shift_axis: int
    ) -> Tensor:
        reshaped = tf.reshape(output, transposed_shape)
        perm = tf.roll(identity_perm, shift=shift_axis, axis=0)
        return tf.transpose(reshaped, perm=perm)

    if isinstance(elems, list):
        if len(elems) > 2:
            raise ValueError("Passing 3 or more tensors as input is not supported.")
        elems_rank = tf.rank(elems[0])
        original_shape = tf.shape(elems[0])
    else:
        elems_rank = tf.rank(elems)
        original_shape = tf.shape(elems)

    # Permutation tensor that does nothing/identity
    identity_perm = tf.range(start=0, limit=elems_rank)
    # Mod the axis param by the rank of the tensor and add 1. To resolve the positive
    # axis value when axis is negative.
    # Create the shift axis. We will roll the identity permutation by this amount to
    # transpose the input
    shift_axis = tf.math.mod(axis, elems_rank) + 1
    # Roll by negative shift axis. For example if
    # axis=0, shift_axis=1, identity_perm=[0, 1, 2]
    # Then transpose_perm = [1, 2, 0]
    transpose_perm = tf.roll(identity_perm, shift=-shift_axis, axis=0)

    # Transpose and reshape
    if isinstance(elems, list):
        reshaped_input = (
            apply_transpose_and_reshape(elems[0]),
            apply_transpose_and_reshape(elems[1]),
        )
    else:
        reshaped_input = apply_transpose_and_reshape(elems)

    # Apply map_fn
    output = tf.map_fn(
        fn=fn,
        elems=reshaped_input,
        parallel_iterations=parallel_iterations,
        swap_memory=swap_memory,
        infer_shape=infer_shape,
        name=name,
        fn_output_signature=fn_output_signature,
    )

    # Undo reshape and transpose
    transposed_shape = tf.gather(original_shape, transpose_perm)
    return apply_undo_transpose_and_reshape(
        output, transposed_shape, identity_perm, shift_axis
    )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class ListRankLayer(BaseLayer):
    """
    Calculate the rank across the axis dimension.

    Example: calculate the rank of items within a query, given the score.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        input_dtype: Optional[str] = None,
        output_dtype: Optional[str] = None,
        sort_order: str = "desc",
        axis: int = 1,
        **kwargs: Any,
    ) -> None:
        """
        Initializes the Listwise Rank layer.

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param sort_order: The order to sort the input tensor by. Defaults to 'desc'
        :param axis: The axis to calculate the rank across. Defaults to 1.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.sort_order = sort_order
        self.axis = axis

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [
            tf.bfloat16,
            tf.float16,
            tf.float32,
            tf.float64,
            tf.uint8,
            tf.int8,
            tf.uint16,
            tf.int16,
            tf.int32,
            tf.int64,
        ]

    @enforce_single_tensor_input
    def _call(self, inputs: Iterable[Tensor], **kwargs: Any) -> Tensor:
        """
        Calculate the rank.

        :param inputs: The iterable tensor for the feature.
        :returns: The new tensor result column.
        """
        return tf.math.add(
            tf.argsort(
                tf.argsort(
                    inputs,
                    axis=self.axis,
                    direction="ASCENDING" if self.sort_order == "asc" else "DESCENDING",
                    stable=True,
                ),
                axis=self.axis,
                stable=True,
            ),
            1,
        )

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the layer.
        Used for saving and loading from a model.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "axis": self.axis,
                "sort_order": self.sort_order,
            }
        )
        return config
